#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "util.h"
#include "cheatsfile.h"

uint32_t smCheckSum(const unsigned char *buf, size_t len) {
	return crc(0, buf, len) ^ 0x5fa27ed6;
}

void usage(void) {
	fprintf(stderr, "usage: sm-crcfix codes\n");
	exit(EXIT_FAILURE);
}

int main(int argc, char *argv[]) {
	FILE *fp;
	smSave_t smSave;
	uint32_t file_checkSum;
	unsigned char *buf;
	const size_t smSaveHeaderSize =
		sizeof(smSave.header.checkSum)    +
		sizeof(smSave.header.dataSize)    +
		sizeof(smSave.header.numOfGames)  +
		sizeof(smSave.header.cheatsTotal) +
		sizeof(smSave.header.codesTotal);

	if (argc < 2)
		usage();

	if ((fp = fopen(argv[1], "rb+")) == NULL)
		goto Error_Exit;

	if (!fgetword(&file_checkSum, fp))
		goto Error_Invalid;
	if (!fgetword(&smSave.header.dataSize, fp))
		goto Error_Invalid;
	if (smSave.header.dataSize < 0)
		goto Error_Invalid;

	buf = malloc(smSave.header.dataSize);
	if (buf == NULL)
		goto Error_Exit;

	if (fseek(fp, smSaveHeaderSize, SEEK_SET))
		goto Error_Exit;
	if ((int32_t)fread(buf, 1, smSave.header.dataSize, fp) < smSave.header.dataSize)
		goto Error_Invalid;

	smSave.header.checkSum = smCheckSum(buf, smSave.header.dataSize);
	if (smSave.header.checkSum == file_checkSum) {
		printf("%s: Checksum OK: %08x\n", argv[1], file_checkSum);
	} else {
		printf("%s: Checksum Fix: %08x -> %08x\n", argv[1], file_checkSum, smSave.header.checkSum);

		if (fseek(fp, 0, SEEK_SET))
			goto Error_Exit;
		fwriteword(smSave.header.checkSum, fp);
	}

	return EXIT_SUCCESS;

Error_Invalid:
	fprintf(stderr, "%s: Not a cheats file or invalid\n", argv[1]);
	exit(EXIT_FAILURE);

Error_Exit:
	fprintf(stderr, "%s: %s\n", argv[1], strerror(errno));
	exit(EXIT_FAILURE);
}
