#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "util.h"
#include "cheatsfile.h"

char *addPrintableChar(char *dst, uint16_t c) {
	/* Xploder uses UTF-32 */
	const char *kanaMap[] = {
		"a","A","i","I","u","U","e","E","o","O",
		"KA","GA","KI","GI","KU","GU","KE","GE","KO","GO",
		"SA","ZA","SI","JI","SU","ZU","SE","ZE","SO","ZO",
		"TA","DA","CHI","JI","tsu","TSU","ZU","TE","DE","TO","DO",
		"NA","NI","NU","NE","NO",
		"HA","BA","PA","HI","BI","PI","FU","BU","PU","HE","BE","PE","HO","BO","PO",
		"MA","MI","MU","ME","MO",
		"ya","YA","yu","YU","yo","YO",
		"RA","RI","RU","RE","RO",
		"wa","WA","I","E","O","N",
		"VU","ka","ke",
		"VA", "VI","VE","VO",
		" ", "-",
	};

	/* "%" */
	if (c == '%' || c == 0xff05) {
		*dst++ = '%';
		*dst++ = '%';
	/* ASCII */
	} else if (c == '\0' || (c >= ' ' && c <= 'z') || c == '{' || c == '}') {
		*dst++ = c;
	/* FULLWIDTH */
	} else if ((c >= 0xff01 && c <= 0xff5a) || c == 0xff5b || c == 0xff5d) {
		*dst++ = c - 0xff01 + '!';
	/* IDEOGRAPHIC SPACE */
	} else if (c == 0x3000) {
		*dst++ = ' ';
	/* HIRAGANA & KATAKANA */
	} else if ((c >= 0x3041 && c <= 0x3096) || (c >= 0x30a1 && c <= 0x30fc)) {
		int o = c - ((c >= 0x30a1) ? 0x30a1 : 0x3041);
		int i = 0;

		while (kanaMap[o][i] != '\0')
			*dst++ = kanaMap[o][i++];
	} else
		*dst++ = '-';

	return dst;
}

void readXPLsave(xplSave_t *xplSave, const char *path) {
	FILE *fp;
	int16_t numOfGames = 0, numOfCodes, numOfCheats;
	int32_t buf;
	char str[256];
	char *ptr;

	if ((fp = fopen(path, "rb")) == NULL) {
		fprintf(stderr, "%s: %s\n", path, strerror(errno));
		exit(EXIT_FAILURE);
	}

	if (!fgetword(&xplSave->header.dataSize, fp))
		goto Error_Invalid;
	if (!fgetword(&xplSave->header.checkSum, fp))
		goto Error_Invalid;

	for (int game_o = 0; ; game_o++, numOfGames++) {
		/* Check EOF */
		if (!fgetword(&buf, fp))
			goto Error_Invalid;

		if ((uint32_t)buf == 0xffffffff)
			break;

		for (ptr = str; ;) {
			ptr = addPrintableChar(ptr, buf);
			/* Check a string terminator */
			if (buf == 0x00000000)
				break;

			if (!fgetword(&buf, fp))
				goto Error_Invalid;
		}

		xplSave->game[game_o].title = malloc(ptr - str + 1);
		if (xplSave->game[game_o].title == NULL)
			goto Error_Exit;
		strcpy(xplSave->game[game_o].title, str);

		if (!fgethalf(&numOfCheats, fp))
			goto Error_Invalid;

		xplSave->game[game_o].numOfCheats = numOfCheats;
		if (numOfCheats > 0) {
			xplSave->game[game_o].cheat = malloc(sizeof(xplSaveCheats_t) * numOfCheats);
			if (xplSave->game[game_o].cheat == NULL)
				goto Error_Exit;

			for (int cheat_o = 0; cheat_o < numOfCheats; cheat_o++) {
				for (ptr = str; ;) {
					if (!fgetword(&buf, fp))
						goto Error_Invalid;

					ptr = addPrintableChar(ptr, buf);
					/* Check a string terminator */
					if (buf == 0x00000000)
						break;
				}

				xplSave->game[game_o].cheat[cheat_o].name = malloc(ptr - str + 1);
				if (xplSave->game[game_o].cheat[cheat_o].name == NULL)
					goto Error_Exit;
				strcpy(xplSave->game[game_o].cheat[cheat_o].name, str);

				if (!fgethalf(&numOfCodes, fp))
					goto Error_Invalid;

				xplSave->game[game_o].cheat[cheat_o].numOfCodes = numOfCodes;
				if (numOfCodes > 0) {
					xplSave->game[game_o].cheat[cheat_o].code = malloc(sizeof(code_t) * numOfCodes);
					if (xplSave->game[game_o].cheat[cheat_o].code == NULL)
						goto Error_Exit;

					for (int code_o = 0; code_o < numOfCodes; code_o++) {
						if (!fgetword(&xplSave->game[game_o].cheat[cheat_o].code[code_o].addr, fp))
							goto Error_Invalid;
						if (!fgetword(&xplSave->game[game_o].cheat[cheat_o].code[code_o].value, fp))
							goto Error_Invalid;
					}
				}
			}
		}
	}

	fclose(fp);

	xplSave->numOfGames = numOfGames;
	return;

Error_Invalid:
	fprintf(stderr, "%s: Not a cheats file or invalid\n", path);
	exit(EXIT_FAILURE);

Error_Exit:
	fprintf(stderr, "%s\n", strerror(errno));
	exit(EXIT_FAILURE);
}

void convertXPL2SM(xplSave_t *xplSave, smSave_t *smSave) {
	int32_t numOfGames, dataSize = 0, cheats_total = 0, codes_total = 0;
	int16_t numOfCheats, numOfCodes;
	size_t len;

	numOfGames = xplSave->numOfGames;
	smSave->game = malloc(sizeof(smSaveGames_t) * numOfGames);
	if (smSave->game == NULL)
		goto Error_Exit;

	for (int game_o = 0; game_o < numOfGames; game_o++) {
		numOfCheats = xplSave->game[game_o].numOfCheats;
		len = strlen(xplSave->game[game_o].title);
		smSave->game[game_o].numOfCheats = numOfCheats;
		smSave->game[game_o].titleLen = len + 1;
		smSave->game[game_o].title = malloc(smSave->game[game_o].titleLen);
		if (smSave->game[game_o].title == NULL)
			goto Error_Exit;
		strcpy(smSave->game[game_o].title, xplSave->game[game_o].title);

		if (numOfCheats > 0) {
			smSave->game[game_o].cheat = malloc(sizeof(smSaveCheats_t) * numOfCheats);
			if (smSave->game[game_o].cheat == NULL)
				goto Error_Exit;

			for (int cheat_o = 0; cheat_o < numOfCheats; cheat_o++) {
				numOfCodes = xplSave->game[game_o].cheat[cheat_o].numOfCodes;
				len = strlen(xplSave->game[game_o].cheat[cheat_o].name);
				smSave->game[game_o].cheat[cheat_o].numOfCodes = numOfCodes;
				smSave->game[game_o].cheat[cheat_o].nameLen = len + 1;
				smSave->game[game_o].cheat[cheat_o].name = malloc(smSave->game[game_o].cheat[cheat_o].nameLen);
				if (smSave->game[game_o].cheat[cheat_o].name == NULL)
					goto Error_Exit;
				strcpy(smSave->game[game_o].cheat[cheat_o].name, xplSave->game[game_o].cheat[cheat_o].name);

				if (numOfCodes > 0) {
					smSave->game[game_o].cheat[cheat_o].code = malloc(sizeof(code_t) * numOfCodes);
					if (smSave->game[game_o].cheat[cheat_o].code == NULL)
						goto Error_Exit;

					for (int code_o = 0; code_o < numOfCodes; code_o++) {
						smSave->game[game_o].cheat[cheat_o].code[code_o] = xplSave->game[game_o].cheat[cheat_o].code[code_o];

						dataSize += sizeof(smSave->game[game_o].cheat[cheat_o].code[code_o].addr);
						dataSize += sizeof(smSave->game[game_o].cheat[cheat_o].code[code_o].value);
						codes_total++;
					}
				}

				dataSize += sizeof(smSave->game[game_o].cheat[cheat_o].numOfCodes);
				dataSize += sizeof(smSave->game[game_o].cheat[cheat_o].nameLen);
				dataSize += smSave->game[game_o].cheat[cheat_o].nameLen;
				cheats_total++;
			}
		}

		dataSize += sizeof(smSave->game[game_o].numOfCheats);
		dataSize += sizeof(smSave->game[game_o].titleLen);
		dataSize += smSave->game[game_o].titleLen;
	}

	smSave->header.checkSum = 0xdeadbeef;	/* Dummy */
	smSave->header.dataSize = dataSize;
	smSave->header.numOfGames = numOfGames;
	smSave->header.cheatsTotal = cheats_total;
	smSave->header.codesTotal = codes_total;

	return;

Error_Exit:
	fprintf(stderr, "%s\n", strerror(errno));
	exit(EXIT_FAILURE);
}

void writeSMsave(smSave_t *smSave, const char *path) {
	FILE *fp;
	int32_t numOfGames, numOfCheats, numOfCodes;

	if ((fp = fopen(path, "wb")) == NULL) {
		fprintf(stderr, "%s: %s\n", path, strerror(errno));
		exit(EXIT_FAILURE);
	}

	fwriteword(smSave->header.checkSum, fp);
	fwriteword(smSave->header.dataSize, fp);
	fwriteword(smSave->header.numOfGames, fp);
	fwriteword(smSave->header.cheatsTotal, fp);
	fwriteword(smSave->header.codesTotal, fp);

	numOfGames = smSave->header.numOfGames;
	for (int game_o = 0; game_o < numOfGames; game_o++) {
		fwritehalf(smSave->game[game_o].numOfCheats, fp);
		fwritebyte(smSave->game[game_o].titleLen, fp);
		fputs(smSave->game[game_o].title, fp);
		fputc('\0', fp);

		numOfCheats = smSave->game[game_o].numOfCheats;
		for (int cheat_o = 0; cheat_o < numOfCheats; cheat_o++) {
			fwritehalf(smSave->game[game_o].cheat[cheat_o].numOfCodes, fp);
			fwritebyte(smSave->game[game_o].cheat[cheat_o].nameLen, fp);
			fputs(smSave->game[game_o].cheat[cheat_o].name, fp);
			fputc('\0', fp);

			numOfCodes = smSave->game[game_o].cheat[cheat_o].numOfCodes;
			for (int code_o = 0; code_o < numOfCodes; code_o++) {
				fwriteword(smSave->game[game_o].cheat[cheat_o].code[code_o].addr, fp);
				fwriteword(smSave->game[game_o].cheat[cheat_o].code[code_o].value, fp);
			}
		}
	}

	fclose(fp);
}

void usage(void) {
	fprintf(stderr, "usage: xpl2sm-conv cheats [codes]\n");
	exit(EXIT_FAILURE);
}

int main(int argc, char *argv[]) {
	xplSave_t xplSave;
	smSave_t smSave;
	const char *output_filename = "codes";

	if (argc < 2)
		usage();

	if (argc > 2)
		output_filename = argv[2];

	memset(&xplSave, 0, sizeof(xplSave));
	memset(&smSave, 0, sizeof(smSave));
	readXPLsave(&xplSave, argv[1]);
	convertXPL2SM(&xplSave, &smSave);
	writeSMsave(&smSave, output_filename);

	return EXIT_SUCCESS;
}
